#version 430 core

#define PI 3.14159265359

layout(local_size_x = 1,
	local_size_y = 1,
	local_size_z = 1) in;

struct VertexStruct {
	vec3 position;
	vec3 normal;
	vec2 uv;
	uint numberOfNeighbors;
	uint offset;
};

layout(std430, binding = 0) buffer VerticesIn {
	VertexStruct vertices[];
}vertIn;

layout(std430, binding = 1) buffer VerticesOut {
	VertexStruct vertices[];
}vertOut;

layout(std430, binding = 2) buffer VertexPointIndicesIn {
	uint vertexPointIndices[];
}vpIndicesIn;

layout(std430, binding = 3) buffer VertexPointIndicesOut {
	uint vertexPointIndices[];
}vpIndicesOut;

layout(std430, binding = 4) buffer EdgePointTableIn {
	uvec4 row[];
}edgePointTableIn;

layout(std430, binding = 5) buffer EdgePointTableOut {
	uvec4 row[];
}edgePointTableOut;

layout(std430, binding = 6) buffer IndicesOut {
	uint indices[];
}indicesOut;

layout(binding = 7) uniform atomic_uint indicesOutCount;

uniform uint vertexCount;
uniform uint edgeCount;
uniform uint subdivisionLevel;


void main(void) {
	uint currentVertex = gl_GlobalInvocationID.x;
	if (currentVertex < vertexCount) {
        // compute new vertex points
		uint numberOfNeighbors = vertIn.vertices[currentVertex].numberOfNeighbors;
		uint offset = vertIn.vertices[currentVertex].offset;

		vec3 neighborPositionSum = vec3(0.0f);
		for (uint i = 0; i < numberOfNeighbors; i++) {
			uint neighbor = vpIndicesIn.vertexPointIndices[offset + i];
			neighborPositionSum += vertIn.vertices[neighbor].position;
		}

		float beta = 1.0f / numberOfNeighbors * (5.0f / 8.0f - pow(3.0f / 8.0f + 1.0f / 4.0f * cos(2.0f*PI / numberOfNeighbors), 2.0f));

		vertOut.vertices[currentVertex].position = (1 - numberOfNeighbors * beta) * vertIn.vertices[currentVertex].position +
			beta * neighborPositionSum;
		vertOut.vertices[currentVertex].normal = vertIn.vertices[currentVertex].normal;
		vertOut.vertices[currentVertex].uv = vertIn.vertices[currentVertex].uv;
		vertOut.vertices[currentVertex].numberOfNeighbors = vertIn.vertices[currentVertex].numberOfNeighbors;
		vertOut.vertices[currentVertex].offset = vertIn.vertices[currentVertex].offset;

        // connnect to adjacent edge vertices
		int found = 0;
		for (uint i = 0; i < edgeCount && found < numberOfNeighbors; i++) {
			if (edgePointTableIn.row[i].x == currentVertex ||
				edgePointTableIn.row[i].y == currentVertex) {
				vpIndicesOut.vertexPointIndices[offset + found] = vertexCount + i;
				found++;
			}
		}
		barrier();
		for (uint i = offset; i < offset + numberOfNeighbors; i++) {
			uint p1 = vpIndicesIn.vertexPointIndices[i];
			for (uint j = i + 1; j < offset + numberOfNeighbors; j++) {
				uint p2 = vpIndicesIn.vertexPointIndices[j];
				for (uint u = vertIn.vertices[p2].offset; u < vertIn.vertices[p2].offset + vertIn.vertices[p2].numberOfNeighbors; u++) {
					if (vpIndicesIn.vertexPointIndices[u] == p1) {;
						uint point1, point2;
						for (uint v = 0; v < edgeCount; v++) {
							if ((edgePointTableIn.row[v].x == currentVertex && edgePointTableIn.row[v].y == p1) ||
								(edgePointTableIn.row[v].x == p1 && edgePointTableIn.row[v].y == currentVertex)) {
								point1 = v + vertexCount;
							} else if ((edgePointTableIn.row[v].x == currentVertex && edgePointTableIn.row[v].y == p2) ||
                                (edgePointTableIn.row[v].x == p2 && edgePointTableIn.row[v].y == currentVertex)) {
								point2 = v + vertexCount;
							}
						}
						uint triangleIdx = atomicCounterIncrement(indicesOutCount);
						indicesOut.indices[triangleIdx * 3 + 0] = currentVertex;
						indicesOut.indices[triangleIdx * 3 + 1] = point1;
						indicesOut.indices[triangleIdx * 3 + 2] = point2;
					}
				}
			}
		}
	} else {
        // compute new edge points
		uint currentEdge = currentVertex - vertexCount;
		uint numberOfAdjacentFaces = 0;
		uint first = edgePointTableIn.row[currentEdge].x;
		uint second = edgePointTableIn.row[currentEdge].y;
		vec3 newPosition = vec3(0.0f);
		vec3 newNormal = vec3(0.0f);
		vec2 newUV = vec2(0.0f);
		if (edgePointTableIn.row[currentEdge].z != first) {
			numberOfAdjacentFaces++;
			newPosition += vertIn.vertices[edgePointTableIn.row[currentEdge].z].position / 8.0f;
			newNormal += vertIn.vertices[edgePointTableIn.row[currentEdge].z].normal / 8.0f;
			newUV += vertIn.vertices[edgePointTableIn.row[currentEdge].z].uv / 8.0f;
		}
		if (edgePointTableIn.row[currentEdge].w != first) {
			numberOfAdjacentFaces++;
			newPosition += vertIn.vertices[edgePointTableIn.row[currentEdge].w].position / 8.0f;
			newNormal += vertIn.vertices[edgePointTableIn.row[currentEdge].w].normal / 8.0f;
			newUV += vertIn.vertices[edgePointTableIn.row[currentEdge].w].uv / 8.0f;
		}

		newPosition += vertIn.vertices[first].position * ((8.0f - numberOfAdjacentFaces) / 16.0f);
		newPosition += vertIn.vertices[second].position * ((8.0f - numberOfAdjacentFaces) / 16.0f);
		newNormal += vertIn.vertices[first].normal * ((8.0f - numberOfAdjacentFaces) / 16.0f);
		newNormal += vertIn.vertices[second].normal * ((8.0f - numberOfAdjacentFaces) / 16.0f);
		newUV += vertIn.vertices[first].uv * ((8.0f - numberOfAdjacentFaces) / 16.0f);
		newUV += vertIn.vertices[second].uv * ((8.0f - numberOfAdjacentFaces) / 16.0f);

		uint offset = vpIndicesIn.vertexPointIndices.length() + currentEdge * 6;
		vertOut.vertices[currentVertex].position = newPosition;
		vertOut.vertices[currentVertex].normal = newNormal;
		vertOut.vertices[currentVertex].uv = newUV;
		vertOut.vertices[currentVertex].numberOfNeighbors = 2 * (numberOfAdjacentFaces + 1);
		vertOut.vertices[currentVertex].offset = offset;

		vpIndicesOut.vertexPointIndices[offset + 0] = first;
		vpIndicesOut.vertexPointIndices[offset + 1] = second;

		barrier();
		if (edgePointTableIn.row[currentEdge].z != first) {
			uint point1, point2;
			for (uint i = 0; i < edgeCount; i++) {
				if ((edgePointTableIn.row[i].x == edgePointTableIn.row[currentEdge].z &&
					edgePointTableIn.row[i].y == first) ||
					(edgePointTableIn.row[i].x == first &&
					edgePointTableIn.row[i].y == edgePointTableIn.row[currentEdge].z)) {
					point1 = vertexCount + i;
				} else if ((edgePointTableIn.row[i].x == edgePointTableIn.row[currentEdge].z &&
					edgePointTableIn.row[i].y == second) ||
					(edgePointTableIn.row[i].x == second &&
						edgePointTableIn.row[i].y == edgePointTableIn.row[currentEdge].z)) {
					point2 = vertexCount + i;
				}
			}
			if (currentVertex < point1 && currentVertex < point2) {
				uint triangleIdx = atomicCounterIncrement(indicesOutCount);
				indicesOut.indices[triangleIdx * 3 + 0] = currentVertex;
				indicesOut.indices[triangleIdx * 3 + 1] = point1;
				indicesOut.indices[triangleIdx * 3 + 2] = point2;
			}
			vpIndicesOut.vertexPointIndices[offset + 2] = point1;
			vpIndicesOut.vertexPointIndices[offset + 3] = point2;
		}
		if (edgePointTableIn.row[currentEdge].w != first) {
			uint point1, point2;
			for (uint i = 0; i < edgeCount; i++) {
				if ((edgePointTableIn.row[i].x == edgePointTableIn.row[currentEdge].w &&
					edgePointTableIn.row[i].y == first) ||
					(edgePointTableIn.row[i].x == first &&
						edgePointTableIn.row[i].y == edgePointTableIn.row[currentEdge].w)) {
					point1 = vertexCount + i;
				}
				else if ((edgePointTableIn.row[i].x == edgePointTableIn.row[currentEdge].w &&
					edgePointTableIn.row[i].y == second) ||
					(edgePointTableIn.row[i].x == second &&
						edgePointTableIn.row[i].y == edgePointTableIn.row[currentEdge].w)) {
					point2 = vertexCount + i;
				}
			}
			if (currentVertex < point1 && currentVertex < point2) {
				uint triangleIdx = atomicCounterIncrement(indicesOutCount);
				indicesOut.indices[triangleIdx * 3 + 0] = currentVertex;
				indicesOut.indices[triangleIdx * 3 + 1] = point1;
				indicesOut.indices[triangleIdx * 3 + 2] = point2;
			}
			vpIndicesOut.vertexPointIndices[offset + 2 * (numberOfAdjacentFaces + 1) - 2] = point1;
			vpIndicesOut.vertexPointIndices[offset + 2 * (numberOfAdjacentFaces + 1) - 1] = point2;
		}
	}
}